/**
  *  \file tok.c
  *  \brief Simple Tokenizer
  *
  *  This is a simple tokenizer for breaking strings apart.
  *  Maybe it's overkill for our simple problems, but at least
  *  it's convenient to have ;-)
  *
  *  \author Stefan Reuther <Streu@gmx.de>
  *
  *  (c) copyright 2001 Stefan Reuther, Thomas Klebes.
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  version 2, as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */

#include <ctype.h>
#include <errno.h>
#include "mfq.h"

/** Maximum size of a token. Static limits suck, but who enters
    100 digit numbers? */
#define TOK_MAX 100

/** The token, for TT_ID */
char the_token[TOK_MAX];

/** Token type. Either a single character, or one of the TT_xxx's */
int token_type;

/** Token value, for int/float */
union TokenVal token_cooked;

/** String we're parsing */
const char* token_ptr;

/** Internal: check if c is a character that can occur in digits. */
static int
IsDigit(unsigned char c)
{
    return isdigit(c) || c == '-' || c == '.';
}

/** Internal: check if c is a character that can occur in identifiers. */
static int
IsWord(unsigned char c)
{
    return isalnum(c);
}

/** Internal: fetch a sequence of characters as classified by func
    into the_token */
static void
getWord(int (*func)(unsigned char))
{
    size_t n = 0;
    while (func(*token_ptr)) {
        if (n < TOK_MAX-1)
            the_token[n++] = *token_ptr;
        ++token_ptr;
    }
    the_token[n] = 0;
}

/** Main entry point: read one token; returns its type.

    \param reading_config parser type. If zero, we're reading
    `mfqX.txt' files: comment character is ';', errors are non-fatal.
    If nonzero, we're reading the config file: comment character is '#',
    errors are fatal. */
int
getToken(int reading_config)
{
    int comment = reading_config ? '#' : ';';
    while (*token_ptr && isspace(*token_ptr))
        ++token_ptr;
    if (*token_ptr == 0 || *token_ptr == comment) {
        return token_type = TT_NULL;
    } else if (IsDigit(*token_ptr)) {
        char* ret;

        getWord(IsDigit);
        /* is it a long? */
        errno = 0;
        token_cooked.l = strtol(the_token, &ret, 0);
        if (errno == 0 && *ret == 0)
            return token_type = TT_INT;

        /* is it a float? */
        errno = 0;
        token_cooked.d = strtod(the_token, &ret);
        if (errno == 0 && *ret == 0)
            return token_type = TT_FLOAT;

        if (reading_config)
            ErrorExit("Syntax error: invalid number: `%s'", the_token);
        return TT_ERROR;
    } else if (IsWord(*token_ptr)) {
        getWord(IsWord);
        return token_type = TT_ID;
    } else {
        return token_type = *token_ptr++;
    }
}
